from typing import List
import pandas as pd
import numpy as np
import danbi as bi

class ZeroBaseMinMaxScaler():
    def __init__(self, scale: List[float] = [-1.0, 1.0], zero_base: bool = False, zero_add: bool = False, same_scale: bool = False):
        # assert (zero_base or not same_scale), "Parameter same_scale can be used only when zero_base is True"
        self._scale = np.array(scale, dtype='float64')
        self._zero_base = zero_base
        self._zero_add = zero_add
        self._same_scale = same_scale
        self._df_columns = None
        self._fits = {}
    
    def store(self) -> dict:
        return {
            "scale": self._scale,
            "zero_base": self._zero_base,
            "zero_add": self._zero_add,
            "same_scale": self._same_scale,
            "df_columns": self._df_columns,
            "fits": self._fits
        }

    def storeFile(self, file: str) -> dict:
        bi.storePickle(self.store(), file)
    
    def restore(self, states):
        self._scale = states["scale"]
        self._zero_base = states["zero_base"]
        self._zero_add = states["zero_add"]
        self._same_scale = states["same_scale"]
        self._df_columns = states["df_columns"]
        self._fits = states["fits"]

    def restoreFile(self, file: str):
        stored_dict = bi.restorePickle(file)
        self.restore(stored_dict)

    def _check_minmax(self, val: List):
        if isinstance(val, list):
            val = np.array(val).astype(np.float64)
        if self._zero_add:
            val = np.append(val, 0.0)
        neg_min = np.nanmin(val)
        pos_max = np.nanmax(val)
        pos_min = neg_max = None

        if self._zero_base:
            if len(val[val >= 0]) > 0:
                pos_min = np.nanmin(val[val >= 0])
            else:
                pos_min = 0
            if len(val[val <= 0]) > 0:
                neg_max = np.nanmax(val[val <= 0])
            else:
                neg_max = 0
            if self._same_scale:
                if np.abs(neg_min) > np.abs(pos_max):
                    pos_max = neg_min * -1
                else:
                    neg_min = pos_max * -1
                if np.abs(neg_max) > np.abs(pos_min):
                    pos_min = neg_max * -1
                else:
                    neg_max = pos_min * -1

        return neg_min, neg_max, pos_min, pos_max

    def fit(self, val: List, field: str = "base", verbos: bool = False):
        neg_min, neg_max, pos_min, pos_max = self._check_minmax(val)
        if verbos:
            print(f"'{field}' min max [neg_min:{neg_min}, neg_max:{neg_max} | pos_min:{pos_min}, pos_max:{pos_max}]")

        if self._zero_base:
            neg_b = neg_max
            neg_w = 0 if (neg_min - neg_b) == 0 else self._scale[0] / (neg_min - neg_b)
            pos_b = pos_min
            pos_w = 0 if (pos_max - pos_b) == 0 else self._scale[1] / (pos_max - pos_b)
            if verbos:
                print(f"   > zerobase is True [neg_b:{neg_b}, neg_w:{neg_w} | pos_b:{pos_b}, pos_w:{pos_w}]")
            self._fits[field] = [neg_b, neg_w, pos_b, pos_w]
        else:
            b = neg_min
            w = (self._scale[1] - self._scale[0]) / (pos_max - neg_min)
            if verbos:
                print(f"   > zerobase is False [b:{b}, w:{w}]")
            self._fits[field] = [b, w]

    def transform(self, val, field="base"):
        if isinstance(val, list):
            val = np.array(val).astype(np.float64)
        if self._zero_base:
            neg_b, neg_w, pos_b, pos_w = self._fits[field]

            val_neg = (val[np.where(val < 0)] - neg_b) * neg_w
            val_neg[val_neg == 0] = -1e-9
            val[val < 0] = val_neg

            val_pos = (val[np.where(val > 0)] - pos_b) * pos_w
            val_pos[val_pos == 0] = 1e-9
            val[val > 0] = val_pos
        else:
            b, w = self._fits[field]
            val = (val - b) * w + self._scale[0]
        return val.round(6)

    def inverse(self, val, field="base"):
        if isinstance(val, list):
            val = np.array(val).astype(np.float64)
        if self._zero_base:
            neg_b, neg_w, pos_b, pos_w = self._fits[field]
            val[val < 0] = val[np.where(val < 0)] / neg_w + neg_b 
            val[val > 0] = val[np.where(val > 0)] / pos_w + pos_b
        else:
            b, w = self._fits[field]
            val = (val - self._scale[0]) / w + b
        return val

    def fitDf(self, df: pd.DataFrame, columns: List[str] = None, groups: List[List[str]] = [[]], verbos=False):
        if columns is None:
            columns = df.select_dtypes(include='number').columns.tolist()
        self._df_columns = columns

        columns = set(columns) - set(sum(groups, []))
        for column in columns:
            self.fit(df[column].astype(np.float64).values, column, verbos)

        for group in groups:
            for idx, item in enumerate(group):
                if idx == 0:
                    temp = df[group].astype(np.float64).values
                    tmp_neg = temp[temp <= 0]
                    tmp_pos = temp[temp >= 0]
                    data = np.array([tmp_neg.min(initial=0), tmp_neg.max(initial=0), tmp_pos.min(initial=0), tmp_pos.max(initial=0)])
                self.fit(data, item, verbos)

    def transformDf(self, df: pd.DataFrame, columns: List[str] = None, inplace: bool = False):
        if columns is None:
            columns = df.select_dtypes(include='number').columns.tolist() if self._df_columns is None else self._df_columns

        transformed = {}
        for column in columns:
            if column in df:
                trans = self.transform(df[column].astype(np.float64).values, column)
                if inplace:
                    df[column] = trans
                else:
                    transformed[column] = trans
        
        if inplace is False:
            return pd.DataFrame(transformed)

    def inverseDf(self, df: pd.DataFrame, columns: List[str] = None):
        if columns is None:
            columns = df.select_dtypes(include='number').columns.tolist()

        transformed = {}
        for column in columns:
            transformed[column] = self.inverse(df[column].values, column)

        return pd.DataFrame(transformed)


def getMinMaxRows(df: pd.DataFrame, cols: List[str] = None):
    """Extract only min/max rows from each column in pandas dataframe.

    Args:
        df (pd.DataFrame): target dataframe.
        cols (List[str], optional): If specified, it is executed only for the specified columns. Defaults to None.

    Returns:
        _type_: _description_
    """
    minmax_rows = set()
    for col in df.columns[[dtype.kind in ["i", "f"] for dtype in df.dtypes]] if cols is None else cols:
        minmax_rows.add(df[col].idxmin())
        minmax_rows.add(df[col].idxmax())
        
    return df.iloc[list(minmax_rows)]



import tensorflow as tf
from tensorflow.keras import layers, saving


@saving.register_keras_serializable(package="danbi.bitensorflow.layers", name="FeatureMasking")
class FeatureMasking(layers.Layer):
    """ A custom Keras layer for feature masking during training.
    Args:
        indices (list or None): A list of indices representing the features to be masked.
            If None or an empty list ([]), no masking is applied, and all features are used.
        retain (bool): If True, the features specified by `indices` are retained and the rest are masked off.
            If False, the specified features are masked off, and the rest are retained.
        axis (int): The axis along which the masking should be applied. Default is the last axis (-1).
        **kwargs: Additional keyword arguments to be passed to the base Layer class.
    Example:
>>> inputs = tf.keras.Input(shape=(10,))
>>> x = FeatureMasking(indices=[0, 2, 4], retain=True)(inputs)
>>> x = tf.keras.layers.Dense(10, activation='relu')(x)
>>> outputs = tf.keras.layers.Dense(1, activation='sigmoid')(x)
>>> model = tf.keras.Model(inputs=inputs, outputs=outputs)
>>> model.compile(optimizer='adam', loss='binary_crossentropy')
    """
    def __init__(self, indices=None, retain=True, axis=-1, **kwargs):
        super(FeatureMasking, self).__init__(**kwargs)
        self._indices = indices
        self._retain = retain
        self._axis = axis
 
    def build(self, input_shape):
        # 입력 shape에 따라 마스크를 생성합니다.
        if self._indices is None or len(self._indices) == 0:
            # 마스크 인덱스가 주어지지 않으면 모든 피쳐를 활성화합니다.
            self.mask = tf.ones(input_shape[self._axis], dtype=tf.bool)
        else:
            self.mask = tf.zeros(input_shape[self._axis], dtype=tf.bool)
            indices = tf.constant(self._indices, dtype=tf.int32)
            self.mask = tf.tensor_scatter_nd_update(self.mask, tf.expand_dims(indices, axis=1), tf.ones_like(indices, dtype=tf.bool))
            if not self._retain:
                self.mask = tf.logical_not(self.mask)
 
    def call(self, inputs, training=None):
        if training:
            # 학습 중에는 마스크를 적용합니다.
            mask = tf.broadcast_to(self.mask, tf.shape(inputs))
            return tf.where(mask, inputs, tf.zeros_like(inputs))
        else:
            # 예측 시에는 모든 피쳐를 사용합니다.
            return inputs
 
    def get_config(self):
        config = super(FeatureMasking, self).get_config()
        config.update({
            "indices": self._indices,
            "retain": self._retain,
            "axis": self._axis,
        })
        return config


import numpy as np
from matplotlib import pyplot, ticker, get_backend, rc, font_manager
from functools import wraps
import re

class EduPlotConf(object):
    def __init__(
        self, dpi=100, figSize=2, figScale=1, lineWidth=1, axisWidth=0.7, markerSize=4,
        font="serif", fontSize=10, textFontSize=11, tickFontSize=9,
        gridParams   = {"linewidth": 0.6, "alpha": 0.3},
        quiverParams = {"angles": "xy", "scale_units": "xy", "scale": 1, "width": 0.007, "headwidth":5, "headaxislength":3},
        axisParam    = {"titlesize":12, "titleweight":"bold", "unicode_minus":False}
    ):
        self.font = font
        self.fontSize = fontSize
        self.textFontSize = textFontSize
        self.dpi = dpi
        self.figSize = np.array([figSize, figSize]) * figScale * 2
        self.lineWidth = lineWidth
        self.axisWidth = axisWidth
        self.markerSize = markerSize
        self.gridParams = gridParams
        self.quiverParams = quiverParams
        self.axisParam = axisParam
        self.tickFontSize = tickFontSize
        
    def set(self):
        rc("font", family=self.font, size=self.fontSize)
        rc("figure", dpi=self.dpi)
        rc("lines", linewidth=self.lineWidth, markersize=self.markerSize)
        rc("axes", titlesize=self.axisParam["titlesize"], titleweight=self.axisParam["titleweight"], unicode_minus=self.axisParam["unicode_minus"])
        rc('xtick', labelsize=self.tickFontSize)
        rc('ytick', labelsize=self.tickFontSize)

    def getFontsList(self, hintRegex=""):
        regex = re.compile(hintRegex, re.IGNORECASE)
        return [{f.name: f.fname} for f in font_manager.fontManager.ttflist if regex.search(f.fname)]


class EduPlot2D(object):
    def __init__(self, conf=EduPlotConf()):
        self._conf = conf
        conf.set()
        self._items = {}
        self.clear()
    
    def _baseSpace(self):
        xticks = self._axis.get_xticks()
        yticks = self._axis.get_yticks()
        dx = xticks[1] - xticks[0]
        dy = yticks[1] - yticks[0]
        base = max(int(min(dx, dy)), 1)   # grid interval is always an integer
        loc = ticker.MultipleLocator(base=base)
        self._axis.xaxis.set_major_locator(loc)
        self._axis.yaxis.set_major_locator(loc)
        self._axis.grid(True, **self._conf.gridParams)
        self._axis.spines["left"].set_linewidth(self._conf.axisWidth)
        self._axis.spines["bottom"].set_linewidth(self._conf.axisWidth)
    
    def _transSpace(self, transMatrix):
        grid_range = 20
        x = np.arange(-grid_range, grid_range+1)
        X_, Y_ = np.meshgrid(x,x)
        I = transMatrix[:,0]
        J = transMatrix[:,1]
        X = I[0]*X_ + J[0]*Y_
        Y = I[1]*X_ + J[1]*Y_
        origin = np.zeros(1)

        # draw grid lines
        for i in range(x.size):
            self._axis.plot(X[:,i], Y[:,i], c="#3D076A", **self._conf.gridParams)
            self._axis.plot(X[i,:], Y[i,:], c="#EAB529", **self._conf.gridParams)
            
        self._axis.spines["left"].set_color("#3D076A")
        self._axis.spines["bottom"].set_color("#EAB529")
        self._axis.spines["left"].set_linewidth(0.3)
        self._axis.spines["bottom"].set_linewidth(0.3)
    
    def _transVector(self, transMatrix, arrays):
        if transMatrix is None:
            return arrays
        result = []
        for item in arrays:
            result.append(transMatrix.dot(item))
        return np.array(result)

    def genSpaceFigure(self, xLim, yLim=None, title=None, transMatrix=None):
        self.genSpace(xLim, yLim, title, transMatrix, False)
        return self._figure

    def genSpace(self, xLim, yLim=None, title=None, transMatrix=None, show=True):
        # input check
        if isinstance(xLim, int): xLim = [-xLim, xLim]
        else: assert len(xLim) == 2, "xLim should have 2 elements list. [-2, 4]"
        if yLim is None: yLim = xLim
        else:
            if isinstance(yLim, int): yLim = [-yLim, yLim]
            else: assert len(yLim) == 2, "xLim should have 2 elements list. [-2, 4]"
        assert xLim[0] <= 0 <= xLim[1], "xLim range should contain 0 elements"
        assert yLim[0] <= 0 <= yLim[1], "yLim range should contain 0 elements"
            
        # create panal with limit
        self._figure, self._axis = pyplot.subplots(figsize=self._conf.figSize)
        if title is not None: self._axis.set_title(title)
        self._axis.set_xlim(xLim)
        self._axis.set_ylim(yLim)
        self._axis.set_aspect("equal")
        
        # generate grid
        if transMatrix is None: self._baseSpace()
        else:
            transMatrix = np.array(transMatrix)
            assert transMatrix.shape == (2,2), "the input matrix must have a shape of (2,2)"
            self._transSpace(transMatrix)
        
        # show x-y axis in the center, hide frames
        self._axis.spines["left"].set_position(("data", 0))
        self._axis.spines["bottom"].set_position(("data", 0))
        self._axis.spines["right"].set_color("none")
        self._axis.spines["top"].set_color("none")
        
        # draw plot
        item = self._items["vectors"]
        for name in item:
            vectors = self._transVector(transMatrix, item[name]["vectors"])
            origins = self._transVector(transMatrix, item[name]["origins"])
            self._axis.quiver(origins[:,0], origins[:,1], vectors[:,0], vectors[:,1],
                              color=item[name]["color"], **self._conf.quiverParams)
            
        item = self._items["functions"]
        for name in item:
            x = np.linspace(xLim[0], xLim[1], item[name]["linespace"])
            y = eval(item[name]["expression"])
            array = self._transVector(transMatrix, [ (i, j) for i, j in zip(x, y)])
            array = np.array(array)
            pyplot.plot(array[:,0], array[:,1], item[name]["color"])
        
        item = self._items["xydata"]
        for name in item:
            pyplot.plot(item[name]["x"], item[name]["y"], item[name]["color"])
        
        item = self._items["markers"]
        for name in item:
            positions = self._transVector(transMatrix, item[name]["positions"])
            pyplot.plot(positions[:,0], positions[:,1], marker=item[name]["marker"], linewidth=0, color=item[name]["color"])
        
        item = self._items["texts"]
        for name in item:
            positions = self._transVector(transMatrix, item[name]["positions"])
            for idx, position in enumerate(positions):
                pyplot.text(position[0], position[1], item[name]["texts"][idx], color=item[name]["color"],
                            horizontalalignment=item[name]["hAlign"], verticalalignment=item[name]["vAlign"], wrap=True, fontsize=self._conf.textFontSize)
        if show: pyplot.show()
        pyplot.close()
    
    def addVector(self, vectors, origins=None, name="vector", color="#0000FF"):
        vectors = np.array(vectors)
        assert vectors.shape[1] == 2, "Each vector should have 2 elements."  
        if origins is not None:
            origins = np.array(origins)
            assert origins.shape[1] == 2, "Each tail should have 2 elements."
        else:
            origins = np.zeros_like(vectors)
        
        nvectors = vectors.shape[0]
        norigins = origins.shape[0]
        if nvectors == 1 and norigins > 1:
            vectors = np.tile(vectors, (norigins, 1))
        elif norigins == 1 and nvectors > 1:
            origins = np.tile(origins, (nvectors, 1))
        else:
            assert origins.shape == vectors.shape, "vectors and tail must have a same shape"
            
        self._items["vectors"][name] = {"vectors":vectors, "origins":origins, "color":color}
    
    def addFunction(self, expX_ForY="x**2", linespace=100, name="function", color="#FF0000"):
        self._items["functions"][name] = {"expression":expX_ForY, "linespace":linespace, "color":color}
        
    def addXYData(self, x, y, name="xydata", color="#0000FF"):
        self._items["xydata"][name] = {"x":x, "y":y, "color":color}
        
    def addMarker(self, positions, marker=".", name="marker", color="#00FF00"):
        positions = np.array(positions)
        self._items["markers"][name] = {"positions":positions, "marker":marker, "color":color}
        
    def addText(self, positions, texts, name="text", color="#000000", hAlign="center", vAlign="bottom"):
        assert len(positions) == len(texts), "positions and texts must have a same size."
        self._items["texts"][name] = {"positions":positions, "texts":texts, "color":color, "hAlign":hAlign, "vAlign":vAlign}
        
    def setConf(self, conf):
        self._conf = conf
        conf.set()
    
    def getConf(self):
        return self._conf
    
    def getFigure(self):
        return self._figure
    
    def clear(self):
        self._items["vectors"] = {}
        self._items["functions"] = {}
        self._items["xydata"] = {}
        self._items["markers"] = {}
        self._items["texts"] = {}


import asyncpg
from .IConnectionManager import IConnectionManager

class ConnMngPsqlAsync(IConnectionManager):
    async def connect(self, **kwargs) -> IConnectionManager:
        try:
            await self.close()
            self._kwargs.update(kwargs)
            
            self._conn_pool = await asyncpg.create_pool(**self._kwargs)
            return self.instance
        except Exception:
            raise
    
    def isConnect(self) -> bool:
        return self._conn_pool is not None
    
    async def close(self, **kwargs) -> None:
        if self._conn_pool is not None:
            await self._conn_pool.close()
    
    async def getConnection(self, **kwargs) -> asyncpg.pool.PoolConnectionProxy:
        try:
            conn = await self._conn_pool.acquire()
            return conn
        except Exception:
            raise
    
    async def releaseConnection(self, conn) -> None:
        try:
            await self._conn_pool.release(conn)
        except Exception:
            raise


import pickle
import bz2

def storePickle(data, file_path: str):
    with bz2.open(file_path, "wb") as f:
        pickle.dump(data, f)

def restorePickle(file_path: str, default: any = {}):
    try:
        with bz2.open(file_path, "rb") as f:
            data = pickle.load(f)
    except:
        data = default
    return data




from jinja2 import Template
from .YAMLConfig import YAMLConfig
from .IMapper import IMapper

class Jinja2Mapper(IMapper):
    def __init__(self, conf_paths: list = None, namespace: str = None, tag = None, base_package: str = None):
        self._conf_paths = conf_paths
        self._mapper = {}
        self._base_package = base_package
        if (namespace is not None) and (tag is not None):
            self.setNamespaceTag(namespace, tag)
    
    def setConfigPaths(self, conf_paths: list = None) -> IMapper:
        self._conf_paths = conf_paths

        return self
    
    def getConfigPaths(self) -> list:
        return self._conf_paths

    def setNamespaceTag(self, namespace: str, tag: str, base_package: str = None) -> None:
        if base_package is not None:
            self._base_package = base_package
        template = YAMLConfig(self._conf_paths, self._base_package)
        template.setCurrent(namespace, tag)
        
        self._mapper = {}
        for config in template.getCurrent():
            for mapper in config["mapper"]:
                self._mapper[mapper["name"]] = mapper["temp"]
    
    def get(self, name: str, values=None, verbose=False) -> str:
        result = Template(self._mapper[name]).render(values=values)
        if verbose:
            print(result)
        return result



from typing import Any, Union, List
import yaml
import os, pkgutil, glob

class YAMLConfig:
    """ Manage various config file with yaml file.
    """
    def __init__(self, conf_paths: List[str] = [], base_package: str = None):
        """
        Args:
            conf_paths (list, optional): file or directory path for yaml config file. Defaults to [].
        """
        assert isinstance(conf_paths, list), "param 'conf_paths' have to a list"
        self._configs = []
        self._current = []
        
        self._first_name, self._first_tag = None, None
        if base_package is None:
            for conf_file in self._parsePaths(conf_paths):
                config = {}
                config["path"] = conf_file
                config["configs"] = []
                with open(conf_file, "r") as data:
                    self._add_config(config, data)
        else:
            for conf_file in conf_paths:
                if conf_file.endswith(".yaml") or conf_file.endswith(".yml"):
                    config = {}
                    config["path"] = base_package + conf_file
                    config["configs"] = []
                    data = pkgutil.get_data(base_package, conf_file)
                    self._add_config(config, data)

        self.setCurrent(self._first_name, self._first_tag)

    def _add_config(self, config, data):
        configs_raw = yaml.safe_load_all(data)
        for data in configs_raw:
            config["configs"].append(data)
            if self._first_name is None:
                self._first_name = data["namespace"]
                self._first_tag = data["tag"]
        self._configs.append(config)

    def _parsePaths(self, conf_paths: list = []) -> list:
        all_paths = []
        for conf_path in conf_paths:
            if os.path.isfile(conf_path):
                all_paths.append(conf_path)
            elif os.path.isdir(conf_path):
                for conf_file in glob.glob(conf_path+"/*.yaml"):
                    all_paths.append(conf_file)
        return all_paths
    
    def getSignatures(self) -> list:
        """
        Returns:
            list: all (namespace, tag, path)
        """
        signatures = []
        for config_meta in self._configs:
            for config in config_meta["configs"]:
                signatures.append((config["namespace"], config["tag"], config_meta["path"]))
        
        return signatures
    
    def getConfig(self, namespace: str, tag: Any) -> dict:
        """Search for settings of config file with the same "name:tag".
        Args:
            namespace (str): namespace of configs
            tag (str): tag/version of configs
        Raises:
            Exception: Occurs when searching for "name:tag" that does not exist.
        Returns:
            dict: all config information.
        """
        result = []
        for config in self._findConfig(namespace, tag)["configs"]:
            result.append(config["config"])
        return result
    
    def _findConfig(self, namespace: str, tag: Any) -> dict:
        result = {"configs": []}
        for config_meta in self._configs:
            for config in config_meta["configs"]:
                if config["namespace"] == namespace and config["tag"] == tag:
                    result["namespace"] = namespace
                    result["tag"] = tag
                    result["configs"].append({
                        "path": config_meta["path"],
                        "config": config
                    })
        
        if len(result["configs"]) == 0:
            raise Exception(f"There is no sutable data. in [namespace: {namespace}, tag: {tag}].")
        else:
            return result
    
    def setCurrent(self, namespace: str, tag: Any):
        """Set the signature ("namespace:tag") to use. 
        Args:
            namespace (str): namespace of config.
            tag (str): tag or version of config.
        """
        self._current = self._findConfig(namespace, tag)

        return self
    
    def getCurrent(self) -> dict:
        """Get all configuration variables existing in the "namespace:tag" specified by the setCurrent() method.
        Returns:
            dict: all configuration variables.
        """
        result = []
        for config in self._current["configs"]:
            result.append(config["config"])
        
        return result
    
    def __genCmdString(self, dot_query: str) -> str:
        command = "config['config']"
        for element in dot_query.split("."):
            try:
                array = element[element.index("["):]
                element = element[:element.index("[")]
            except Exception:
                array = None
                pass

            command += f"['{element}']"
            if array is not None:
                command += array
        return command
    
    def getValue(self, dot_query: str) -> str:
        """The value of the configuration variables in the "namespace:tag" specified by the setCurrent() method.
        Args:
            dot_query (str): variable position using dot notation.
            ex) a.b[1].c
        Returns:
            str: the value of the position.
        """
        result = []
        for config in self._current["configs"]:
            result.append(eval(self.__genCmdString(dot_query)))
        return result
    
    def setValue(self, dot_query: str, value: Union[tuple, dict, list]) -> None:
        """Set the value at the location of the variable using dot notation.
        Args:
            dot_query (str): variable position using dot notation.
            value (str): The setting value you want to set.
        """
        for config in self._current["configs"]:
            exec(f"{self.__genCmdString(dot_query)}='{value}'")
    
    def persist(self) -> None:
        """save the current configurations to the source file.
        """
        for current in self._current["configs"]:
            for config in self._configs:
                if config["path"] == current["path"]:
                    with open(config["path"], 'w') as file:
                        yaml.dump_all(config['configs'], file)



class DotDict:
    """
    A dictionary subclass that supports dot notation access to its keys.
    Nested dictionaries are also converted to DotDict instances.
    """

    def __init__(self, data):
        """
        Initialize the DotDict with a dictionary.
        
        Args:
            data (dict): The dictionary to convert.
        """
        for key, value in data.items():
            if isinstance(value, dict):
                # Recursively convert nested dictionaries
                setattr(self, key, DotDict(value))
            else:
                setattr(self, key, value)

    def __getattr__(self, attr):
        """
        Retrieve an attribute. Raises AttributeError if not found.
        
        Args:
            attr (str): The attribute name.
        
        Returns:
            The attribute value.
        
        Raises:
            AttributeError: If the attribute does not exist.
        """
        try:
            return self.__dict__[attr]
        except KeyError:
            raise AttributeError(f"'DotDict' object has no attribute '{attr}'")

    def __setattr__(self, key, value):
        """
        Set an attribute. If the value is a dictionary, convert it to DotDict.
        
        Args:
            key (str): The attribute name.
            value: The value to set.
        """
        if isinstance(value, dict):
            value = DotDict(value)
        self.__dict__[key] = value

    def to_dict(self):
        """
        Convert the DotDict back to a regular dictionary.
        
        Returns:
            dict: The converted dictionary.
        """
        result = {}
        for key, value in self.__dict__.items():
            if isinstance(value, DotDict):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result

    def __repr__(self):
        """
        Official string representation of DotDict.
        
        Returns:
            str: The string representation.
        """
        return f"DotDict({self.to_dict()})"

    def __iter__(self):
        """
        Make DotDict iterable by iterating over its keys.
        
        Returns:
            Iterator over the keys of the DotDict.
        """
        return iter(self.__dict__)

    def __len__(self):
        """
        Return the number of items in the DotDict.
        
        Returns:
            int: Number of key-value pairs.
        """
        return len(self.__dict__)

    def keys(self):
        """
        Return a new view of the DotDict's keys.
        
        Returns:
            dict_keys: A view object displaying a list of all the keys.
        """
        return self.__dict__.keys()

    def values(self):
        """
        Return a new view of the DotDict's values.
        
        Returns:
            dict_values: A view object displaying a list of all the values.
        """
        return self.__dict__.values()

    def items(self):
        """
        Return a new view of the DotDict's items (key, value pairs).
        
        Returns:
            dict_items: A view object displaying a list of the DotDict's key-value pairs.
        """
        return self.__dict__.items()



import abc
import collections
import itertools
from typing import Any, Dict, Tuple, List, Union
import pandas as pd
from multiprocessing import Lock

from .IConnectionManager import IConnectionManager
from ..mapping.IMapper import IMapper

class IDB(abc.ABC):
    def __new__(self, manager: IConnectionManager, mapper: IMapper):
        self._manager = manager
        self._mapper = mapper
        self._lock_q = Lock()
        self._lock_qr = Lock()
        self._lock_qp = Lock()
        self._lock_qpr = Lock()
        self._lock_e = Lock()
        self._lock_er = Lock()
        self._lock_em = Lock()
        self._lock_emr = Lock()

        if not hasattr(self, 'instance'):
            self.instance = super(IDB, self).__new__(self)
        return self.instance

    def _pyformat2psql(self, query: str, named_args: Dict[str, Any]) -> Tuple[str, List[Any]]:
        positional_generator = itertools.count(1)
        positional_map = collections.defaultdict(lambda: '${}'.format(next(positional_generator)))
        formatted_query = query % positional_map
        positional_items = sorted(
            positional_map.items(),
            key=lambda item: int(item[1].replace('$', '')),
        )
        positional_args = [named_args[named_arg] for named_arg, _ in positional_items]
        return formatted_query, positional_args

    def getManager(self) -> IConnectionManager:
        """
        Returns:
            IConnectionManager: danbi's database connection control manager.
        """
        return self._manager
    
    def getMapper(self) -> IMapper:
        """
        Returns:
            IMapper: danbi's Jinja2Mapper control manager.
        """
        return self._mapper
    
    @abc.abstractmethod
    def query(self, mapper_name: str, values: Union[dict, tuple] = None, print_sql: bool = False) -> list:
        """query with Jinja2Mapper's query key.

        Args:
            mapper_name (str): key value of yaml file of Jinja2Mapper.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.

        Returns:
            list: result of query
        """
        ...
    
    @abc.abstractmethod
    def queryRaw(self, raw_sql: str, values: tuple = None) -> list:
        """query with raw sql.

        Args:
            raw_sql (str): query sql from raw string.
            values (tuple, optional): parameter values for sql. Defaults to None.

        Returns:
            list: result of query
        """
        ...
    
    @abc.abstractmethod
    def queryPandas(self, mapper_name: str, values: Union[dict, tuple] = None, dtype: dict = None, print_sql: bool = False) -> pd.DataFrame:
        """query with Jinja2Mapper's query key. and the result type change to pandas dataframe.

        Args:
            mapper_name (str): key value of yaml file of Jinja2Mapper.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.
            dtype (dict, optional): pandas column's data type. The name of the column depends on the contents of the select clause of the query.  Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe.
        """
        ...
    
    @abc.abstractmethod
    def queryPandasRaw(self, raw_sql: str, values: Union[dict, tuple] = None, dtype: dict = None) -> pd.DataFrame:
        """query with raw sql query by raw_sql. and the result type change to pandas dataframe.

        Args:
            raw_sql (str): query sql from raw string.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.
            dtype (dict, optional): pandas column's data type. The name of the column depends on the contents of the select clause of the query. Defaults to None.

        Returns:
            pd.DataFrame: pandas dataframe.
        """
        ...
    
    @abc.abstractmethod
    def execute(self, mapper_name: str, values: Union[dict, tuple] = None, print_sql: bool = False) -> int:
        """all sqls without a result table.

        Args:
            mapper_name (str): key value of yaml file of Jinja2Mapper.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.

        Returns:
            int: The number of results executed.
        """
        ...
    
    @abc.abstractmethod
    def executeRaw(self, raw_sql: str, values: Union[dict, tuple] = None) -> int:
        """all sqls without a result table with raw sql string.

        Args:
            raw_sql (str): query sql from raw string.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.

        Returns:
            int: The number of results executed.
        """
        ...
    
    @abc.abstractmethod
    def executeMany(self, mapper_name: str, values: Union[dict, tuple] = None, print_sql: bool = False) -> int:
        """When inserting a lot of data in batches with Jinja2Mapper's query key.

        Args:
            mapper_name (str): key value of yaml file of Jinja2Mapper.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.

        Returns:
            int: The number of results executed.
        """
        ...
    
    @abc.abstractmethod
    def executeManyRaw(self, raw_sql: str, values: Union[dict, tuple] = None) -> int:
        """hen inserting a lot of data in batches with raw sql string.

        Args:
            raw_sql (str): query sql from raw string.
            values (Union[dict, tuple], optional): parameter values for sql and mapper. Defaults to None.

        Returns:
            int: The number of results executed.
        """
        ...



from danbi import Jinja2Mapper
from danbi.database import ConnMngPsqlAsync, DBPsqlAsync
from danbi.database import ConnMngPsql, DBPsql, IDB

def usePgDBMapper(user: str, password: str, host: str, port: int, database: str, pool_min: int, pool_max: int, mappers: list, namespace: str, tag, base_package: str = None) -> IDB:
    psql = ConnMngPsql(
        user=user,
        password=password,
        host=host,
        port=port,
        database=database
    ).connect(minconn=pool_min, maxconn=pool_max)
    mapper = Jinja2Mapper(mappers, namespace, tag, base_package)
    db = DBPsql(psql, mapper)

    return db

async def usePgDBMapperAsync(user: str, password: str, host: str, port: int, database: str, pool_min: int, pool_max: int, mappers: list, namespace: str, tag, base_package: str = None) -> IDB:
    psql = await ConnMngPsqlAsync(
        user=user,
        password=password,
        host=host,
        port=port,
        database=database
    ).connect(min_size=pool_min, max_size=pool_max)
    mapper = Jinja2Mapper(mappers,namespace, tag, base_package)
    db = DBPsqlAsync(psql, mapper)

    return db




from .IConnectionManager import IConnectionManager
from .ConnMngPsql import ConnMngPsql
from .IDB import IDB
from .DBPsql import DBPsql
from .ConnMngPsqlAsync import ConnMngPsqlAsync
from .DBPsqlAsync import DBPsqlAsync
from .factory import (
    usePgDBMapper, usePgDBMapperAsync
)

from ..mapping import Jinja2Mapper
psql: IDB = DBPsql(ConnMngPsql(), Jinja2Mapper())

def setPsql(user: str, password: str, host: str, port: int, database: str, pool_min: int, pool_max: int, mappers: list, namespace: str, tag, base_package: str = None):
    if psql.getMapper().getConfigPaths() is None:
        psql.getMapper().setConfigPaths(mappers).setNamespaceTag(namespace, tag, base_package)
    if not psql.getManager().isConnect():
        psql.getManager().connect(
            user=user,
            password=password,
            host=host,
            port=port,
            database=database,
            minconn=pool_min,
            maxconn=pool_max,
        )



import numpy as np
import pandas as pd

class WithNumpyPrint:
    def __init__(self, precision: int = 3):
        self._precision = precision
        
    def __enter__(self):
        self.original_options = np.get_printoptions()
        np.set_printoptions(linewidth=1000, threshold=np.inf, precision=self._precision, suppress=True)
        
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        np.set_printoptions(**self.original_options)

class WithPandasPrint:
    def __init__(self, precision=5, width=0, rows=None, columns=None):
        self._width = width
        self._precision = precision
        self._rows = rows
        self._columns = columns
        self._origins = {}

    def __enter__(self):
        self._origins["width"] = pd.get_option("display.width")
        pd.set_option("display.width", self._width)
        
        self._origins["precision"] = pd.get_option("display.precision")
        pd.set_option("display.precision", self._precision)

        self._origins["rows"] = pd.get_option("display.max_rows")
        pd.set_option("display.max_rows", self._rows)

        self._origins["columns"] = pd.get_option("display.max_columns")
        pd.set_option("display.max_columns", self._columns)

    def __exit__(self, exc_type, exc_value, traceback):
        pd.set_option("display.width", self._origins["width"])
        pd.set_option("display.precision", self._origins["precision"])
        pd.set_option("display.max_rows", self._origins["rows"])
        pd.set_option("display.max_columns", self._origins["columns"])



import pandas as pd
import numpy as np
import datetime
from typing import List, Tuple, Any, Union

from bokeh.io import curdoc, show
from bokeh.plotting import figure, show, curdoc
from bokeh.layouts import gridplot, column
from bokeh.models import ColumnDataSource, HoverTool, CrosshairTool, Span
from bokeh.models.formatters import DatetimeTickFormatter

from .plot_chart import tools, COLORSET


def showBokeh(fig: figure):
    show(fig)


def showPandas(df: pd.DataFrame, xlist: Union[str, List[str]], ylist: Union[str, List[Union[str, List[str]]]], typeList: Union[str, List[List[str]]] = "line", width: int = 1500, height: int = 400, col_dir: bool = True, sync_axis: str = "xy",
               graph_widths: Union[int, List[int]] = 1, line_colors: List[str] = None, hover=True, **options):
    cds = ColumnDataSource(df)
    """
    example: showPandas(df, "a", ["b", ["c", "d", "e"]], col_dir=False)
    """
    toolbar_location = options.get("toolbar_location", "above")
    legend_location = options.get("legend_location", "top_right")
    local_tools = options.get("tools", tools)
    title = options.get("title", "Learning History")
    title_size = options.get("title_size", "11pt")
    format_datetime = options.get("format_datetime", "%Y-%m-%d")
    
    len_y = len(ylist)
    if col_dir:
        width = int(width/len_y)
    else:
        height = int(height/len_y)
    
    if isinstance(xlist, str):
        xlist = [xlist for _ in range(len_y)]
    if isinstance(typeList, str):
        typeList = [typeList for _ in range(len_y)]
    
    plot_figures = []
    for idx, (xname, ynames) in enumerate(zip(xlist, ylist)):
        fig = figure(width=width, height=height, title=" & ".join(ynames), tools=local_tools, toolbar_location=toolbar_location)

        if isinstance(graph_widths, int):
            graph_widths = [graph_widths for _ in range(len(ynames))]
        else:
            if len(graph_widths) < len(ynames):
                for _ in range(len(graph_widths), len(ynames)):
                    graph_widths.append(graph_widths[-1])
        tooltips = [(f"{xname}", f"@{xname}")]
        types = typeList[idx]
        if isinstance(types, str):
            types = [types for _ in range(len(ynames))]
        for idx, name in enumerate(ynames):
            if types[idx] == "line":
                fig_each = fig.line(x=xname, y=name, source=cds, line_width=graph_widths[idx], color=COLORSET[idx % 53], legend_label=name, alpha=0.8, muted_alpha=0.02)
            elif types[idx] == "scatter":
                fig_each = fig.scatter(x=xname, y=name, source=cds, size=graph_widths[idx], color=COLORSET[idx % 53], legend_label=name, alpha=0.8, muted_alpha=0.02)
            
            if idx == 0:
                fig_base = fig_each
            tooltips.append((f"{name}", f"@{name}"))
        fig.title.text_font_size = title_size
        fig.grid.grid_line_alpha = 0.3
        fig.legend.border_line_alpha = 0
        fig.legend.background_fill_alpha = 0
        fig.legend.click_policy = "mute"
        fig.legend.location = legend_location
        if isinstance(cds.data[xname][0], datetime.date) or isinstance(cds.data[xname][0], np.datetime64):
            fig.xaxis.formatter = DatetimeTickFormatter(years=[format_datetime], months=[format_datetime], days=[format_datetime])

        plot_figures.append(fig)
        
        if hover:
            fig.add_tools(HoverTool(
                tooltips=tooltips,
                mode="vline",
                renderers=[fig_base]
            ))
    
    if sync_axis is not None:
        for plot1 in plot_figures:
            for plot2 in plot_figures:
                if "x" in sync_axis:
                    plot1.x_range = plot2.x_range
                if "y" in sync_axis:
                    plot1.y_range = plot2.y_range

    chart_rows = gridplot([plot_figures] if col_dir else [[plot] for plot in plot_figures])
    curdoc().add_root(chart_rows)
    show(chart_rows)


def showAsRows(plots: List[figure], sync_axis: str = None, width: int = None):
    wHair = Span(dimension="width", line_dash="dotted", line_width=0.8)
    hHair = Span(dimension="height", line_dash="dotted", line_width=0.8)
    
    if sync_axis is not None:
        for plot1 in plots:
            plot1.add_tools(CrosshairTool(overlay=[wHair, hHair], line_alpha=0.5))
            if width is not None:
                plot1.width = width
            for plot2 in plots:
                if "x" in sync_axis:
                    plot1.x_range = plot2.x_range
                if "y" in sync_axis:
                    plot1.y_range = plot2.y_range
    
    chart_rows = gridplot([[plot] for plot in plots])
    curdoc().add_root(chart_rows)
    show(chart_rows)


def showAsGrid(plots: List[List[Union[figure, List[figure]]]], sync_axis: str = None, width: int = None):
    wHair = Span(dimension="width", line_dash="dotted", line_width=0.8)
    hHair = Span(dimension="height", line_dash="dotted", line_width=0.8)
    
    if sync_axis is not None:
        plots_flat = []
        for items in sum(plots, []):
            if isinstance(items, list):
                for item in items:
                    plots_flat.append(item)
            else:
                plots_flat.append(items)
        for plot1 in plots_flat:
            plot1.add_tools(CrosshairTool(overlay=[wHair, hHair], line_alpha=0.3))
            if width is not None:
                plot1.width = width
            for plot2 in plots_flat:
                if "x" in sync_axis:
                    plot1.x_range = plot2.x_range
                if "y" in sync_axis:
                    plot1.y_range = plot2.y_range
    
    grid = []
    for rows in plots:
        row = []
        for cols in rows:
            if isinstance(cols, list):
                col = column(cols)
                row.append(col)
            else:
                row.append(cols)
        grid.append(row)
    
    chart_rows = gridplot(grid)
    curdoc().add_root(chart_rows)
    show(chart_rows)






import os, pkgutil, inspect
from .IPlugin import IPlugin
from danbi import utils

class PluginManager:
    def __init__(self, **kwargs):
        self._plugins = {}
        self._kwargs = kwargs
    
    def addPackagePath(self, path: str):
        self._discover_plugins(path)

        return self
    
    def addPackage(self, package_regex_name: str):
        packages_info = utils.infoInstalledPackage(package_regex_name)
        for package in packages_info:
            self._discover_plugins(package[0])
        
        return self

    def getPlugins(self) -> list:
        return list(self._plugins.keys())
    
    def _discover_plugins(self, package: str) -> None:
        imported_package = __import__(package, fromlist=['blah'])
        for _, pluginname, ispkg in pkgutil.iter_modules(imported_package.__path__, imported_package.__name__ + '.'):
            if not ispkg:
                plugin_module = __import__(pluginname, fromlist=['blah'])
                clazz_members = inspect.getmembers(plugin_module, inspect.isclass)
                for (_, clazz) in clazz_members:
                    if issubclass(clazz, IPlugin) & (clazz is not IPlugin):
                        full_name = f'{clazz.__module__}.{clazz.__name__}'
                        self._plugins[full_name] = [None, False, clazz]
        
        all_current_paths = []
        if isinstance(imported_package.__path__, str):
            all_current_paths.append(imported_package.__path__)
        else:
            all_current_paths.extend([x for x in imported_package.__path__])
        
        seen_paths = []
        for pkg_path in all_current_paths:
            if (pkg_path not in seen_paths) and (not pkg_path.endswith("__")):
                seen_paths.append(pkg_path)
                child_pkgs = [p for p in os.listdir(pkg_path) if os.path.isdir(os.path.join(pkg_path, p))]
                for child_pkg in child_pkgs:
                    if child_pkg.startswith(".") or child_pkg.startswith("__"):
                        continue
                    self._discover_plugins(package + '.' + child_pkg)
    
    def plug(self, target: str = None) -> None:
        if target is None:
            for fullname, plugin in self._plugins.items():
                if not plugin[1]:
                    plugin[0] = plugin[2](fullname)
                    plugin[0].plug(**self._kwargs)
                    plugin[1] = True
        else:
            plugin = self._plugins[target]
            if not plugin[1]:
                plugin[0] = plugin[2](target)
                plugin[0].plug(**self._kwargs)
                plugin[1] = True
    
    def unplug(self, target: str = None) -> None:
        if target is None:
            for key, plugin in self._plugins.items():
                if plugin[1]:
                    plugin[0].unplug(**self._kwargs)
                    del plugin[0]
                    self._plugins[key] = [None, False, plugin[1]]
        else:
            plugin = self._plugins[target]
            if plugin[1]:
                plugin[0].unplug(**self._kwargs)
                del plugin[0]
                self._plugins[target] = [None, False, plugin[1]]



from typing import List, Callable
import psutil
import pandas as pd
import ray

class ActorPool:
    """Deploy as many actors as the number of mounts to ray cluster and create a pool manager for manage them.
    """
    def __init__(self, ActorClass: ray.actor.ActorClass, amount: int):
        """Deploy as many actors as the number of mounts to ray cluster and create a pool manager for manage them.

        Args:
            ActorClass (ray.actor.ActorClass): ray's actor class.
            amount (int): amount of actors. in general, use only the number of cpu cores. 
        """
        actors = [ActorClass.remote() for _ in range(amount)]
        self._actor_pool = ray.util.ActorPool(actors)
    
    def get(self, method_name: str, batch_vals: list) -> list:
        """Run at ray cluster with registered actor.

        Args:
            method_name (str): actor's method name of string what you want to run.
            batch_vals (list): data list for multi actors.

        Returns:
            list: Results of processing by multiple actors with the input data list.
        """
        results = self._actor_pool.map(lambda a, v: getattr(a, method_name).remote(*v), batch_vals)
        
        return [x for x in results]
    
    def getPandas(self, method_name: str, batch_vals: list, columns: list, dropna: bool = False) -> pd.DataFrame:
        """It's same as get(). but the result is pandas DataFrame.

        Args:
            method_name (str): actor's method name of string what you want to run.
            batch_vals (list): data list for multi actors.
            columns (list): results column names.
            dropna (bool, optional): Except when there is None among the results. Defaults to False.

        Returns:
            pd.DataFrame: Results of processing by multiple actors with the input data list.
        """
        results = self.get(method_name, batch_vals)
        df = pd.DataFrame(results, columns=columns)
        
        return df.dropna() if dropna else df


def runTasks(func_ray: Callable, vals: List, func_callback: Callable = None, chunk: int = None, vervos: bool = True):
    if chunk is None:
        chunk = int(psutil.cpu_count() * 0.9)
    
    ray_refs = []
    cnt_total = len(vals)
    for idx, val in enumerate(vals):
        if vervos:
            print(f"\r{idx+1}/{cnt_total} ({round((idx+1)/cnt_total*100)}%) {len(ray_refs)} {' '*20}", end="")
        if isinstance(val, (list, tuple)):
            ray_refs.append(func_ray.remote(*val))
        else:
            ray_refs.append(func_ray.remote(val))
        if idx % chunk == 0:
            while len(ray_refs) > int(chunk * 0.5):
                done_id, ray_refs = ray.wait(ray_refs)
                if func_callback is not None:
                    func_callback(*ray.get(done_id)[0])
    while len(ray_refs):
        done_id, ray_refs = ray.wait(ray_refs)
        if func_callback:
            func_callback(*ray.get(done_id)[0])



import pandas as pd
from pandas.api.types import is_datetime64_any_dtype
import numpy as np
import datetime
from typing import List, Tuple, Any, Union, Dict
from sklearn.preprocessing import MinMaxScaler

from bokeh.models import ColumnDataSource, Range1d, LinearAxis, Span, BoxAnnotation
from bokeh.models import HoverTool,WheelZoomTool, PanTool, ResetTool, CrosshairTool, BoxSelectTool, BoxZoomTool, SaveTool
from bokeh.models.formatters import NumeralTickFormatter, DatetimeTickFormatter
from bokeh.plotting import figure, output_notebook

tools = [PanTool(), WheelZoomTool(), ResetTool(), CrosshairTool(line_alpha=0.5), BoxSelectTool(), BoxZoomTool(), SaveTool()]
COLORSET = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#13005A', '#222831', '#3A1078', '#EA8FEA', '#6A2C70', '#5254a3', '#8ca252', '#bd9e39', '#ad494a', '#a55194', '#6baed6', '#fd8d3c', '#74c476', '#9e9ac8', '#969696', '#9c9ede', '#cedb9c', '#e7cb94', '#e7969c', '#de9ed6', '#c6dbef', '#fdd0a2', '#c7e9c0', '#dadaeb', '#d9d9d9', '#1b9e77', '#d95f02', '#7570b3', '#e7298a', '#66a61e', '#e6ab02', '#a6761d', '#666666', '#b3e2cd', '#fdcdac', '#cbd5e8', '#f4cae4', '#e6f5c9', '#fff2ae', '#f1e2cc', '#cccccc', '#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462', '#b3de69', '#fccde5', '#d9d9d9', '#bc80bd', '#ccebc5', '#ffed6f']


def setJupyterEnable():
    output_notebook()


def getDataSource(df: pd.DataFrame) -> ColumnDataSource:
    return ColumnDataSource(df)


def getFigure(width: int, height: int, title: str, datetime: bool = True, tools: List = tools, toolbar_location: str = "above"):
    if datetime:
        fig = figure(width=width, height=height, title=title, tools=tools, toolbar_location=toolbar_location, x_axis_type='datetime')
    else:
        fig = figure(width=width, height=height, title=title, tools=tools, toolbar_location=toolbar_location)
    
    return fig


def setFigureStyle(fig: figure, tooltips: List = None, formatters: tuple = None, base_plot: Any = None, mode: str = "vline", **options):
    fig.title.text_font_size = options.get("title_font_size", "12pt")
    fig.grid.grid_line_alpha = options.get("grid_line_alpha", 0.5)
    fig.yaxis.formatter = NumeralTickFormatter(format="0,0[.]00")
    fig.xaxis.formatter = DatetimeTickFormatter(years="%Y-%m", months="%Y-%m", days="%m%d")

    fig.legend.border_line_alpha = 0
    fig.legend.background_fill_alpha = 0
    fig.legend.padding = options.get("legend_padding", -7)
    fig.legend.spacing = options.get("legend_spacing", -4)
    fig.legend.click_policy = "mute"
    fig.legend.location = options.get("legend_location", "top_left")
    fig.legend.label_text_font_size = options.get("legend_font_size", "9pt")
    
    if tooltips and formatters and base_plot:
        fig.add_tools(HoverTool(
            tooltips=tooltips,
            formatters=formatters,
            mode="vline",
            renderers=[base_plot]
        ))
        fig.toolbar.active_inspect = None


def setLine(fig: figure, src: Union[ColumnDataSource, pd.DataFrame], x: str, y: str, color: str, width: int = 1.2, legend_label: str = None, alpha: int = 0.7, muted_alpha: int = 0.05, y_ref: int = None, y_range: tuple = None, extra_y_name: str = None, extra_y_range: tuple = None):
    if isinstance(src, pd.DataFrame):
        src = ColumnDataSource(src)
    
    if legend_label is None:
        legend_label = y
    
    if extra_y_name is not None:
        line = fig.line(source=src, x=x, y=y, color=color, line_width=width, legend_label=legend_label, alpha=alpha, muted_alpha=muted_alpha, y_range_name=extra_y_name)
        fig.circle(x=x, y=y, source=src, radius=0, y_range_name=extra_y_name)
        if extra_y_range is not None:
            fig.extra_y_ranges.update({extra_y_name: Range1d(*extra_y_range)})
            fig.add_layout(LinearAxis(y_range_name=extra_y_name), "right")
        if y_ref is not None:
            fig.renderers.extend([Span(location=y_ref, dimension='width', line_color="slategray", line_width=1, y_range_name=extra_y_name)])
    else:
        line = fig.line(source=src, x=x, y=y, color=color, line_width=width, legend_label=legend_label, alpha=alpha, muted_alpha=muted_alpha)
        fig.circle(x=x, y=y, source=src, radius=0)
        if y_ref is not None:
            fig.renderers.extend([Span(location=y_ref, dimension='width', line_color="slategray", line_width=1)])
    if y_range:
        fig.y_range = Range1d(*y_range)
    
    return line


def setBar(fig: figure, src: Union[ColumnDataSource, pd.DataFrame], x: str, y: str, color: str, width: int = 1.2, legend_label: str = None, alpha: int = 0.7, muted_alpha: int = 0.05):
    if isinstance(src, pd.DataFrame):
        src = ColumnDataSource(src)
    
    if legend_label is None:
        legend_label = y
    
    bar = fig.vbar(source=src, x=x, top=y, color=color, line_width=width, legend_label=legend_label, alpha=alpha, muted_alpha=muted_alpha)
    
    return bar


def setMarke(fig: figure, src: Union[ColumnDataSource, pd.DataFrame], x: str, y: str, color: str, size: int = 10, marker: str = "triangle", legend_label: str = None, alpha: int = 0.5, muted_alpha: int = 0.05):
    if isinstance(src, pd.DataFrame):
        src = ColumnDataSource(src)
    
    if legend_label is None:
        legend_label = y
    
    fig.scatter(x=x, y=y, source=src, size=size, color=color, marker=marker, legend_label=legend_label, alpha=alpha, muted_alpha=muted_alpha)
    
    return fig


def setAreaBox(fig: figure, color: str, alpha: int = 0.5, top: float = None, bottom: float = None, left: float = None, right: float = None):
    fig.add_layout(BoxAnnotation(top=top, bottom=bottom, left=left, right=right, fill_alpha=alpha, fill_color=color))
    
    return fig


def setHVLines(fig, hlines: List[float] = [], vlines: List[float] = [], color: str = "#E90064"):
    extend_lines = []
    for hline in hlines:
        extend_lines.append(Span(location=hline, dimension='width', line_color=color, line_width=0.9, line_alpha=0.6, line_dash='dashed'))
    for vline in vlines:
        extend_lines.append(Span(location=vline, dimension='height', line_color=color, line_width=0.9, line_alpha=0.6, line_dash='dashed'))

    fig.renderers.extend(extend_lines)
    
    return fig


def plotTimeseries(df, x: str, ylist: List[str], height: int = 300, width: int = 1600, hlines: List[float] = [], vlines: List[float] = [], title: str = "Scaled Timeseries", time: bool = True, scale: bool = False, legend: Tuple = {}, hover_time: bool = False):
    df_plot = df[[x] + ylist].copy()
    if hover_time:
        tooltips = [(x, "@"+x+"{%F %T}"), ("index", "@index{0,0}")]
    else:
        tooltips = [(x, "@"+x+"{%F}"), ("index", "@index{0,0}")]
    formatters = {"@"+x: "datetime"}
    if time and not is_datetime64_any_dtype(df_plot[x]):
        df_plot[x] = pd.to_datetime(df_plot[x], utc=True)
    if scale:
        scaler = MinMaxScaler()
        df_plot[ylist] = scaler.fit_transform(df_plot[ylist])
    
    ds = getDataSource(df_plot)
    fig = getFigure(width, height, title)
    for idx, name in enumerate(ylist):
        legend_label = legend[name] if name in legend else name
        if idx == 0:
            base = setLine(fig, ds, x, name, COLORSET[idx], legend_label=legend_label)
        else:
            setLine(fig, ds, x, name, COLORSET[idx], legend_label=legend_label)
        tooltips.append((name, "@"+name+"{0,0[.]00}"))
    
    extend_lines = []
    for hline in hlines:
        extend_lines.append(Span(location=hline, dimension='width', line_color='#E90064', line_width=0.9, line_alpha=0.6, line_dash='dashed'))
    for vline in vlines:
        extend_lines.append(Span(location=vline, dimension='height', line_color='#E90064', line_width=0.9, line_alpha=0.6, line_dash='dashed'))

    fig.renderers.extend(extend_lines)
    
    setFigureStyle(fig, tooltips, formatters, base)
    
    return fig


def plotSeries(datas: Union[Dict, pd.DataFrame], width: int = 1600, height: int = 300, style="line", hlines: List[float] = [], vlines: List[float] = [], title: str = "Series"):
    if isinstance(datas, pd.DataFrame):
        datas = datas.to_dict("list")
    
    fig = getFigure(width, height, title)
    tooltips, formatters = [], {"@time": "datetime"}
    
    cds_dict = {}
    for idx, (key, value) in enumerate(datas.items()):
        cds_dict[f"idx_{key}"] = np.arange(len(value)) + (0e-5 * idx) if style == "vbar" else np.arange(len(value))
        cds_dict[f"{key}"] = value
    tooltips.append(("index", "@idx_"+next(iter(datas))+"{0,0[.]00}"))
    cds = ColumnDataSource(data=cds_dict)

    for idx, key in enumerate(datas):
        if style == "vbar":
            base = fig.vbar(source=cds, x=f"idx_{key}", top=key, width=0.3, color=COLORSET[idx], alpha=0.01, muted_alpha=0, legend_label=key)
        else:
            base = fig.line(source=cds, x=f"idx_{key}", y=key, line_width=0.8, color=COLORSET[idx], alpha=0.8, muted_alpha=0, legend_label=key)
        tooltips.append((key, "@"+key+"{0,0[.]00}"))
    
    extend_lines = []
    for hline in hlines:
        extend_lines.append(Span(location=hline, dimension='width', line_color='#E90064', line_width=0.9, line_alpha=0.6, line_dash='dashed'))
    for vline in vlines:
        extend_lines.append(Span(location=vline, dimension='height', line_color='#E90064', line_width=0.9, line_alpha=0.6, line_dash='dashed'))

    fig.renderers.extend(extend_lines)
        
    setFigureStyle(fig, tooltips, formatters, base_plot=base)
    
    return fig



